/*
  ============================================================================================
  	
	phi-X.c
	by Philipp Bulling, 2012
	
	Metronome with speed trainer 
	Target device: ATmega8
	
	Version: 1.1
	
	For detailed information, visit:
    philaudio.wordpress.com
	
  ============================================================================================
*/

#define XTAL 8000000
#define F_CPU XTAL

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdlib.h>
#include "LCD_display.h"
#include "HardwareInit.h"

// ===========================================================================================
// Function Prototypes:
uint8_t getRotarySwitchState(void);
uint8_t adc(uint8_t channel);
void menu(uint8_t menuCounter);
uint16_t calcLinBPM(void);

// ===========================================================================================
// Global Variables:
volatile uint8_t i = 0;
volatile uint8_t scale = 1;
volatile uint8_t currentBeat;
volatile uint8_t clickOn = 0; 
volatile uint16_t j = 0; 
volatile uint16_t fromBPM = 120; 
volatile uint16_t toBPM = 120; 
volatile uint8_t bars = 8;
volatile uint8_t beatsPerBar = 1;
volatile uint32_t beatCounter = 0;
volatile uint8_t oldADCValue = 120;

// ===========================================================================================
// Interrupts:
ISR(TIMER2_OVF_vect)
{
	// Every time, the TCNT1 has reached TOP, a new value of the sine 
	// is set to OCR2. Once the index i has reached the end of one 
	// sine-period, it is reset to 0.
	
	if (clickOn != 0)
	{
		OCR2 = sinePeriod[i];	// threshold for the PWM 
		if (i==255)	i=0;		// one sine-period has 256 values (index 0...255)
		else i += scale;		// Skip some values to get higher frequency.
		PORTB |= (1<<PB2);		// LED on	
		
		if (j == 1200)			// Length of the played sound
		{
			j=0;					
			clickOn = 0;	
		}			
		else j++;		
	}	
	else
		PORTB &= ~(1<<PB2);		// LED off		
}

ISR(TIMER1_OVF_vect)
{
	//TCNT1 = 34286;			// = 1 Hz = 1 sec = 60 bpm
								// = 2^16 - F_CPU / Timer Prescaler

	// Change speed:
	TCNT1 = 65536 - F_CPU / 256.0 * 60.0 / calcLinBPM();
			
	clickOn = 1;
	
	// First beat accent (play click with higher frequency):
	if (currentBeat == getRotarySwitchState())
	{
		scale = 18;
		currentBeat = 0;
	}			
	else 
	{
		scale = 14;
	}			
	
	currentBeat++;
	
	// Linear speed increasing until toBPM is reached.
	if (beatCounter < (bars * beatsPerBar))
		beatCounter++;
}	


// ===========================================================================================
// Main-Function:
int main(void)
{
	uint8_t str[3];
	uint8_t oldRotarySwitchState = getRotarySwitchState();
	currentBeat = getRotarySwitchState();
	uint8_t menuCounter = 0;
	
	hardwareInit();
	
	// Pre-output on LCD:
	lcd_write(0, 0, "phi-X Metronome");
	lcd_write(1, 0, "Philipp Bulling");
	while (PINB & (1<<PB1));
	lcd_send(COMMAND, LCD_CLEAR);
	
	lcd_write(0, 0, "From|To  |Bars|#");
	lcd_write(1, 0, "    |    |    | ");
	dtostrf(fromBPM, 3, 0, str);
	lcd_write(1, 1, str);
	dtostrf(toBPM, 3, 0, str);
	lcd_write(1, 6, str);
	dtostrf(bars, 3, 0, str);
	lcd_write(1, 11, str);
	
	// Start Timer
	sei();
    	
	// =======================================================================================
	// Main-Loop:
    while(1)
    {	
		// ================================================
		// Debug
		dtostrf(calcLinBPM(), 3, 0, str);
		lcd_write(0, 22, str);
		
		dtostrf(beatCounter, 3, 0, str);
		lcd_write(1, 19, str);
		
		dtostrf(bars*getRotarySwitchState(), 3, 0, str);
		lcd_write(1, 22, str);
		// ================================================
		
		// Show current beat on LCD:
		dtostrf(currentBeat, 1, 0, str);
		lcd_write(1, 15, str);
		
		// Menu key pressed?
		if (~PINB & (1<<PB1))
		{
			// Jump to next menu:
			menuCounter++;
			// Key debouncing:
			_delay_ms(500);			
		}			
			
		// Menu function:
		menu(menuCounter);	
		if (menuCounter == 4) menuCounter = 0;
				
		// Check, if Rotary Switch State has changed:
		if (oldRotarySwitchState != getRotarySwitchState()) 
		{
			// If so, reset beat counter:
			currentBeat = getRotarySwitchState();
			oldRotarySwitchState = getRotarySwitchState();
		}	
    }
}


// ===========================================================================================
// Function Definitions:
uint8_t getRotarySwitchState(void)
{
	// Returns the state of the 4-state rotary switch
	// (values from 1...4).
	
	uint8_t rotarySwitch;
	uint8_t count = 0;
	
	// Rotary Switch at PORTC1-3:
	rotarySwitch = ~(PINC>>1) & 0b000111;
	
	if (rotarySwitch == 0)
	{
		count = 1;
	}		
	else
	{
		count = 2;
		while ((rotarySwitch >>= 1) != 0) count++;
	}
	return count;
}

uint8_t adc(uint8_t channel)
{
	// AD conversion:
	ADMUX = channel & 0x07;			// single ended channel 0..7
	ADMUX |= (1<<REFS0);			// AVCC as reference
	ADMUX |= (1<<ADLAR);			// left adjust result
	
	ADCSRA |= (1<<ADSC);			// start conversion
	while (ADCSRA & (1<<ADSC));		// wait for ADC ready
	
	return ADCH;					// 8-bit precision
}

void menu(uint8_t menuCounter)
{
	// Menu: Lets the user adjust start-tempo, end-tempo
	// and the number of bars inbetween.
	
	uint8_t str[3];
	
	switch (menuCounter)
	{
		case 1:		cli();								// Stop playing sound
					PORTB |= (1<<PB2);					// LED on
					lcd_write(1, 0, ">");
					// Check, if user wants to change the value:
					if (oldADCValue != adc(0))
					{
						oldADCValue = adc(0);
						fromBPM = oldADCValue + 45;
					}						
					dtostrf(fromBPM, 3, 0, str);
					lcd_write(1, 1, str);
					break;
		case 2:		lcd_write(1, 0, " ");
					lcd_write(1, 5, ">");
					// Check, if user wants to change the value:
					if (oldADCValue != adc(0))
					{
						oldADCValue = adc(0);
						toBPM = oldADCValue + 45;
					}						
					dtostrf(toBPM, 3, 0, str);
					lcd_write(1, 6, str);
					break;
		case 3:		lcd_write(1, 0, " ");
					lcd_write(1, 5, " ");
					lcd_write(1, 10, ">");
					// Check, if user wants to change the value:
					if (oldADCValue != adc(0))
					{
						oldADCValue = adc(0);
						bars = (oldADCValue>>1) + 1;	// 7-bit precision (1...128)
					}						
					dtostrf(bars, 3, 0, str);
					lcd_write(1, 11, str);
					break;
		case 4:		lcd_write(1, 0, " ");
					lcd_write(1, 5, " ");
					lcd_write(1, 10, " ");
					currentBeat = getRotarySwitchState();
					beatsPerBar = getRotarySwitchState();
					beatCounter = 0;
					sei();								// Restart playing sound
					break;
		default:	break;
	}
}	

uint16_t calcLinBPM(void)
{
	// Returns linear interpolation, depending on the 
	// current beat, between start tempo and end tempo.
	return ((int16_t)(toBPM - fromBPM) / (float)(bars * beatsPerBar) * beatCounter) + fromBPM;
}